# Written by Robert C. Kirby
# Copyright 2005 by The University of Chicago
# Distributed under the LGPL license
# This work is partially supported by the US Department of Energy
# under award number DE-FG02-04ER25650

import jacobi, shapes, math

"""Principal expansion functions as defined by Karniadakis & Sherwin.
The main point of entry to this module is the function
make_expansions( shape , n ) which takes the simplex shape and the
polynomal degree n and returns the list of expansion functions up to
that degree."""

def psitilde_a( i , z ):
    return jacobi.eval_jacobi( 0 , 0 , i , z )

def psitilde_b( i , j , z ):
    if i == 0:
        return jacobi.eval_jacobi( 1 , 0 , j , z )
    else:
        return (0.5 * ( 1 - z )) ** i * jacobi.eval_jacobi( 2*i+1 , 0 , j , z )

def psitilde_c( i , j , k , z ):
    if i + j == 0:
        return jacobi.eval_jacobi( 2 , 0 , k , z )
    else:
        return ( 0.5 * (1-z))**(i+j) \
               * jacobi.eval_jacobi( 2*(i+j+1), 0 , k , z )

# coordinate changes from Karniadakis & Sherwin

def eta_triangle( xi ):
    ( xi1 , xi2 ) = xi
    if xi2 == 1.:
        eta1 = -1.0
    else:
        eta1 = 2.0 * ( 1. + xi1 ) / ( 1. - xi2 ) - 1
    eta2 = xi2 
    return eta1 , eta2

def xi_triangle( eta ):
    eta1, eta2 = eta
    xi1 = 0.5 * (1. + eta1) * (1. - eta2) - 1.
    xi2 = eta2
    return xi1,xi2

def eta_tetrahedron( xi ):
    xi1,xi2,xi3 = xi
    if xi2 + xi3 == 0.:
        eta1 = 1.
    else:
        eta1 = -2. * ( 1. + xi1 ) / (xi2 + xi3) - 1.
    if xi3 == 1.:
        eta2 = -1.
    else:
        eta2 = 2. * (1. + xi2) / (1. - xi3 ) - 1.
    eta3 = xi3
    return eta1,eta2,eta3


def xi_tetrahedron( eta ):
    eta1,eta2,eta3 = eta

    xi1 = 0.25 * ( 1. + eta1 ) * ( 1. - eta2 ) * ( 1. - eta3 ) - 1.
    xi2 = 0.5 * ( 1. + eta2 ) * ( 1. - eta3 ) - 1.
    xi3 = eta3

    return xi1,xi2,xi3

coord_changes = { shapes.TRIANGLE: eta_triangle , \
                  shapes.TETRAHEDRON: eta_tetrahedron }

inverse_coord_changes = { shapes.TRIANGLE: xi_triangle , \
                          shapes.TETRAHEDRON: xi_tetrahedron }

def make_coordinate_change( shape ):
    """Maps from reference domain to rectangular reference domain."""
    global coord_changes
    if coord_changes.has_key( shape ):
        return coord_changes[shape]
    else:
        raise RuntimeError, "Can't collapse coordinates"

def make_inverse_coordinate_change( shape ):
    """Maps from rectangular reference domain to reference domain."""
    global inverse_coord_changes
    if inverse_coord_changes.has_key( shape ):
        return inverse_coord_changes[shape]
    else:
        raise RuntimeError, "Can't collapse coordinates"

def phi_line( index , xi ):
    (p,) = index
    (eta,) = xi
    return psitilde_a( p , eta )

def phi_triangle( index , xi ):
    p,q = index
    eta1,eta2 = eta_triangle( xi )

    alpha = psitilde_a( p , eta1 )
    beta = psitilde_b( p , q , eta2 )

    return alpha * beta

def phi_tetrahedron( index , xi ):
    p,q,r = index
    eta1,eta2,eta3 = eta_tetrahedron( xi )

    alpha = psitilde_a( p , eta1 )
    beta = psitilde_b( p , q , eta2 )
    gamma = psitilde_c( p , q , r , eta3 )

    return alpha * beta * gamma

class ExpansionFunction( object ):
    def __init__( self , indices , phi , alpha ):
        self.indices, self.phi,self.alpha = indices , phi , alpha
        return
    def __call__( self , x ):
        if len( x ) != len( self.indices ):
            raise RuntimeError, "Illegal number of coordinates"
        return self.alpha * self.phi( self.indices , x )

class PhiLine( ExpansionFunction ):
    def __init__( self , i ):
        ExpansionFunction.__init__( self , \
                                    (i , ) , \
                                    phi_line , \
                                    math.sqrt(1.0*i+0.5) )

class PhiTriangle( ExpansionFunction ):
    def __init__( self , i , j ):
        ExpansionFunction.__init__( self , \
                                    ( i,j ) , \
                                    phi_triangle, \
                                    math.sqrt( (i+0.5)*(i+j+1.0) ) )
        return

class PhiTetrahedron( ExpansionFunction ):
    def __init__( self , i , j , k ):
        ExpansionFunction.__init__( self , (i,j,k) , \
                                    phi_tetrahedron , \
                                    math.sqrt( (i+0.5)*(i+j+1.0)*(i+j+k+1.5)))
        return



def make_phis_line( n ):
    return [ PhiLine( i ) \
             for i in range(0,n+1) ]

def make_phis_triangle( n ):
    return [ PhiTriangle( k - i , i ) \
             for k in range( 0 , n + 1 ) \
             for i in range( 0 , k + 1 ) ]

def make_phis_tetrahedron( n ):
    return [ PhiTetrahedron( k - i - j , j , i ) \
             for k in range( 0 , n + 1 ) \
             for i in range( 0 , k + 1 ) \
             for j in range( 0 , k - i + 1 ) ]

make_phis = { shapes.LINE : make_phis_line , \
              shapes.TRIANGLE : make_phis_triangle , \
              shapes.TETRAHEDRON : make_phis_tetrahedron }

def make_expansion( shape , n ):
    """Returns the orthogonal expansion basis on a given shape
    for polynomials of degree n."""
    global make_phis
    try:
        return make_phis[shape]( n )
    except:
        raise ShapeError, "expansions.make_expansion: Illegal shape"



